program nbc;

{$APPTYPE CONSOLE}
{$UNDEF CAN_DOWNLOAD}

uses
{$IFNDEF FPC}
{$DEFINE CAN_DOWNLOAD}
  FastMM4 in '..\FastMM4.pas',
  FastMM4Messages in '..\FastMM4Messages.pas',
  FastMove in '..\FastMove.pas',
  uSpirit in '..\bricktools\uSpirit.pas',
  FantomSpirit in '..\bricktools\FantomSpirit.pas',
{$ENDIF}
  Classes,
  SysUtils,
  Math,
  uCmdLineUtils in '..\uCmdLineUtils.pas',
  ParamUtils in '..\ParamUtils.pas',
  uVersionInfo in '..\uVersionInfo.pas',
  uNQCLexer in '..\uNQCLexer.pas',
  mwGenericLex in '..\mwGenericLex.pas',
  uCommonUtils in '..\uCommonUtils.pas',
  uNXTClasses in 'uNXTClasses.pas',
  uPreprocess in 'uPreprocess.pas',
  uNXTConstants in 'uNXTConstants.pas',
  uRPGComp in 'uRPGComp.pas',
  uTiny in 'uTiny.pas',
  P10Build in 'P10Build.pas',
  Parser10 in 'Parser10.pas';

{$IFDEF WIN32}
{$R *.RES}
{$ENDIF}

{$IFDEF CAN_DOWNLOAD}
var
  BC : TBrickComm;

function BrickComm : TBrickComm;
begin
  if not Assigned(BC) then
  begin
    BC := TFantomSpirit.Create();
  end;
  Result := BC;
end;
{$ENDIF}

procedure PrintUsage;
begin
  PrintVersion;
  WriteLn('Syntax: ' + progName + ' [options] filename [options]');
  WriteLn('');
  WriteLn('   -T=<target>: target must be NXT');
{$IFDEF CAN_DOWNLOAD}
  WriteLn('   -S=<portname>: specify port name (COMn or usb), resource name, or alias');
  WriteLn('   -BT: use bluetooth');
  Writeln('   -d: download program');
  Writeln('   -b: treat input file as a binary file (don''t compile it)');
  Writeln('   -q: quiet');
{$ENDIF}
//  Writeln('   -D=<sym>: define macro <sym>');
//  Writeln('   -U=<sym>: undefine macro <sym>');
  Writeln('   -x: decompile program');
  Writeln('   -Z[1|2]: turn on compiler optimizations');
  Writeln('   -O=<outfile> : specify output file');
  Writeln('   -E=<filename> : write compiler errors to <filename>');
  Writeln('   -I=<path>: search <path> for include files');
  Writeln('   -L=<filename> : generate code listing to <filename>');
  Writeln('   -Y=<filename> : generate symbol table to <filename>');
  Writeln('   -help : display command line options');
  // compiler also takes an undocumented "nxt name" parameter which is
  // used to tell the compiler what the downloaded program should be called
  // on the NXT: -N=<nxtname>
end;

var
  sIn, sOut : TMemoryStream;
  tmpIncDirs : TStringList;
  C : TRXEProgram;
  TC : TTinyComp;
  RC : TRPGComp;
  F : TextFile;
{$IFDEF CAN_DOWNLOAD}
  theType : TNXTFileType;
{$ENDIF}
  Filename : string;
  nxtName : string;
  TheErrorCode : Byte;
  i, optLevel : integer;
  incDirs : string;
  bNXCErrors : boolean;
//  bOptimize : boolean;

procedure WriteSymbolTable(C : TRXEProgram);
var
  dir, logFilename : string;
begin
  if ParamSwitch('-Y', False) then
  begin
    logFilename := ParamValue('-Y', False);
    dir := ExtractFilePath(logFilename);
    if dir <> '' then
      ForceDirectories(dir);
    C.SymbolTable.SaveToFile(logFilename);
  end;
end;

procedure Decompile(const aFilename : string);
var
  slDump : TStrings;
  D : TRXEDumper;
  i : integer;
begin
  // decompile the .rxe file
  slDump := TStringList.Create;
  try
    D := TRXEDumper.Create;
    try
      D.LoadFromFile(aFilename);
      D.Decompile(slDump);
    finally
      D.Free;
    end;
    if ParamSwitch('-O', False) then
    begin
      // write the contents of slDump to the file
      slDump.SaveToFile(ParamValue('-O', False));
    end
    else
    begin
      // write the contents of slDump to the output
      for i := 0 to slDump.Count - 1 do
        WriteLn(slDump[i]);
    end;
  finally
    slDump.Free;
  end;
end;

function GetDefaultPath : string;
begin
//  Result := ExtractFilePath(ParamStr(0));
  Result := IncludeTrailingPathDelimiter(GetCurrentDir);
end;

function GetCurrentFilename : string;
var
  ext : string;
begin
  ext := ExtractFileExt(Filename);
  Result := ChangeFileExt(nxtName, ext);
  // add a path if there isn't one already
  if ExtractFilename(Result) = Result then
    Result := GetDefaultPath + Result;
end;

begin
  TheErrorCode := 0;
//  bOptimize    := False;

{$IFDEF FPC}
  VerCompanyName      := 'JoCar Consulting';
  VerFileDescription  := '';
  VerFileVersion      := '1.0.1.15';
  VerInternalName     := 'NBC';
  VerLegalCopyright   := 'Copyright (c) 2006, John Hansen';
  VerOriginalFileName := 'NBC';
  VerProductName      := 'Next Byte Codes Compiler';
  VerProductVersion   := '1.0.1.b15';
  VerComments         := '';
{$ENDIF}

try
  if ParamSwitch('-help', False) then
  begin
    PrintUsage;
    Exit;
  end;

  if ParamCount = 0 then
  begin
    PrintUsageError;
    Exit;
  end;

  Filename := getFilename;
  if Trim(Filename) = '' then
  begin
    PrintUsageError;
    Exit;
  end;

  if ParamSwitch('-O', False) then
    nxtName := ParamValue('-O', False)
  else if ParamSwitch('-N', False) then
    nxtName := ParamValue('-N', False)
  else
    nxtName := Filename;

  optLevel := 0;
  if ParamSwitch('-Z', False) or ParamSwitch('-Z2', False) then
    optLevel := 2
  else if ParamSwitch('-Z1', False) then
    optLevel := 1;


{$IFDEF CAN_DOWNLOAD}
  BrickComm.BrickType := rtNXT;
  if ParamSwitch('-S') then
  begin
    BrickComm.Port := ParamValue('-S');
  end
  else
    BrickComm.Port := 'usb'; // if no port is specified then default to usb
  BrickComm.UseBluetooth := ParamSwitch('-BT');
{$ENDIF}

  if ParamSwitch('-x', False) then
  begin
    Decompile(Filename);
  end
  else
  begin
    setErrorOutputFile(F);
    try
      sIn := TMemoryStream.Create;
      try
        sIn.LoadFromFile(Filename);
        if ParamSwitch('-b', False) and ParamSwitch('-d', False) then
        begin
{$IFDEF CAN_DOWNLOAD}
          // just download the already compiled binary file
          if not BrickComm.IsOpen then
            BrickComm.Open;
          theType := NameToNXTFileType(Filename);
          if LowerCase(ExtractFileExt(Filename)) = '.rpg' then
            theType := nftOther;
          BrickComm.NXTDownloadStream(sIn, Filename, theType);
{$ENDIF}
        end
        else
        begin
          tmpIncDirs := TStringList.Create;
          try
            if ParamSwitch('-I', False) then
            begin
              incDirs := ParamValue('-I', False);
              // does the path contain ';'?  If so parse
              i := Pos(';', incDirs);
              while i > 0 do begin
                tmpIncDirs.Add(Copy(incDirs, 1, i-1));
                Delete(incDirs, 1, i);
                i := Pos(';', incDirs);
              end;
              tmpIncDirs.Add(incDirs);
            end;
            if LowerCase(ExtractFileExt(Filename)) = '.npg' then
            begin
              // RPG compiler
              RC := TRPGComp.Create;
              try
                RC.CurrentFile := GetCurrentFilename;
                try
                  RC.Parse(sIn);
                  sOut := TMemoryStream.Create;
                  try
                    if RC.SaveToStream(sOut) then
                    begin
{$IFDEF CAN_DOWNLOAD}
                      if ParamSwitch('-d', False) then
                      begin
                        // download the compiled code to the brick
                        if not BrickComm.IsOpen then
                          BrickComm.Open;
                        BrickComm.NXTDownloadStream(sOut, ChangeFileExt(nxtName, '.rpg'), nftOther);
                      end;
{$ENDIF}
                      if ParamSwitch('-O', False) then
                        sOut.SaveToFile(nxtName);
                    end;
                  finally
                    sOut.Free;
                  end;
                finally
                  // write compiler messages to output
                  for i := 0 to RC.CompilerMessages.Count - 1 do
                    WriteLn(F, RC.CompilerMessages[i]);
                end;
              finally
                RC.Free;
              end;
            end
            else
            begin
              bNXCErrors := False;
              if LowerCase(ExtractFileExt(Filename)) = '.nxc' then
              begin
                TC := TTinyComp.Create;
                try
                  TC.IncludeDirs.AddStrings(tmpIncDirs);
                  TC.CurrentFile := GetCurrentFilename;
                  try
                    TC.Parse(sIn);
                    sIn.Clear;
                    TC.NBCSource.SaveToStream(sIn);
                    optLevel := Max(optLevel, 1);
//                    TC.NBCSource.SaveToFile('__'+ChangeFileExt(Filename, '.nbc'));
                    sIn.Position := 0;
                  finally
                    // write compiler messages to output
                    for i := 0 to TC.CompilerMessages.Count - 1 do
                      WriteLn(F, TC.CompilerMessages[i]);
                  end;
                  bNXCErrors := TC.CompilerMessages.Count > 0;
                finally
                  TC.Free;
                end;
              end;
              if not bNXCErrors then
              begin
                // compile the nbc file
                C := TRXEProgram.Create;
                try
                  C.ReturnRequiredInSubroutine := True;
                  C.OptimizeLevel := optLevel;
                  try
                    C.IncludeDirs.AddStrings(tmpIncDirs);
                    C.CurrentFile := GetCurrentFilename;
                    C.Parse(sIn);
                    sOut := TMemoryStream.Create;
                    try
                      if C.SaveToStream(sOut) then
                      begin
                        WriteCompilerOutput(C.CompilerOutput);
                        WriteSymbolTable(C);
{$IFDEF CAN_DOWNLOAD}
                        if ParamSwitch('-d', False) then
                        begin
                          // download the compiled code to the brick
                          if not BrickComm.IsOpen then
                            BrickComm.Open;
                          BrickComm.NXTDownloadStream(sOut, ChangeFileExt(nxtName, '.rxe'), nftProgram);
                        end;
{$ENDIF}
                        if ParamSwitch('-O', False) then
                          sOut.SaveToFile(nxtName);
                      end;
                    finally
                      sOut.Free;
                    end;
                  finally
                    // write compiler messages to output
                    for i := 0 to C.CompilerMessages.Count - 1 do
                      WriteLn(F, C.CompilerMessages[i]);
                  end;
                finally
                  C.Free;
                end;
              end;
            end;
          finally
            tmpIncDirs.Free;
          end;
        end;
      finally
        sIn.Free;
      end;
    finally
      CloseFile(F);
    end;
  end;
except
  TheErrorCode := 1;
end;

{$IFDEF CAN_DOWNLOAD}
BC.Free;
{$ENDIF}

if TheErrorCode <> 0 then
  Halt(TheErrorCode);

end.
